<?php

namespace Drupal\whatlinkshere\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\node\NodeTypeInterface;
use Drupal\Core\Batch\BatchBuilder;

/**
 * Provides a Drush command to scan for node references.
 */
class WhatlinkshereCommands extends DrushCommands {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a WhatlinkshereCommands object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager) {
    $this->entityTypeManager = $entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager')
    );
  }

  /**
   * Finds references to nodes.
   *
   * @command whatlinkshere:scan
   * @aliases wlhscan
   * @param string|null $nids
   *   A comma-separated list of Node IDs to restrict the scan (optional).
   * @option batch-size
   *   The number of nodes to process in each batch. Default is 50.
   *
   * @usage whatlinkshere:scan
   *   Scan references for all content.
   *
   * @usage whatlinkshere:scan 1,2,3
   *   Scan references for specific node IDs 1, 2, and 3.
   *
   * @usage whatlinkshere:scan --batch-size=100
   *   Scan all content in batches of 100 nodes.
   */
  public function scan($nids = NULL, $options = ['batch-size' => 100]) {
    $batch_size = (int) $options['batch-size'];
    $storage = $this->entityTypeManager->getStorage('node');

    if ($nids) {
      $node_ids = array_map('trim', explode(',', $nids));
      self::processNodes($node_ids, $storage);
    }
    else {
      $query = \Drupal::entityQuery('node')->accessCheck(FALSE);
      $node_ids = $query->execute();

      if (empty($node_ids)) {
        $this->logger()->warning('No nodes found to scan.');
        return;
      }

      $batch_builder = (new BatchBuilder())
        ->setTitle(t('Scanning node references'))
        ->setInitMessage(t('Initialising node reference scan.'))
        ->setProgressMessage(t('Processing batch @current out of @total.'))
        ->setErrorMessage(t('Node reference scan encountered an error.'));

      $chunks = array_chunk($node_ids, $batch_size);
      foreach ($chunks as $chunk) {
        $batch_builder->addOperation([
          self::class, 'processNodesBatch'
        ], [$chunk]);
      }

      batch_set($batch_builder->toArray());
      drush_backend_batch_process();
    }
  }

  /**
   * Processes a batch of nodes and outputs progress.
   */
  public static function processNodesBatch(array $node_ids, &$context) {
    $storage = \Drupal::entityTypeManager()->getStorage('node');
    $context['results']['processed'] = isset($context['results']['processed']) ? $context['results']['processed'] : 0;

    foreach ($node_ids as $nid) {
      $node = $storage->load($nid);
      if ($node) {
        $references = self::findReferences($node);
        if (!empty($references)) {
          \Drupal::logger('whatlinkshere')->notice("Node $nid is referenced by: " . implode(', ', $references));
        }
        else {
          \Drupal::logger('whatlinkshere')->notice("Node $nid has no references.");
        }
      }

      // Update progress message.
      $context['results']['processed']++;
      $context['message'] = t('Processed @count nodes...', ['@count' => $context['results']['processed']]);
    }
  }

  /**
   * Finds nodes that reference a given node.
   *
   * @param \Drupal\node\Entity\Node $node
   *   The node entity.
   *
   * @return array
   *   An array of referencing node IDs.
   */
  private static function findReferences($node) {
    $references = [];
    $entity_type_manager = \Drupal::service('entity_type.manager');
    $node_type = $entity_type_manager->getStorage('node_type')->load($node->bundle());

    if (!$node_type instanceof NodeTypeInterface) {
      return $references;
    }

    $field_definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions('node', $node->bundle());

    foreach ($field_definitions as $field_name => $definition) {
      if ($definition instanceof FieldDefinitionInterface && $definition->getType() === 'entity_reference' && $definition->getSetting('target_type') === 'node') {
        $query = \Drupal::entityQuery('node')
          ->condition($field_name, $node->id())
          ->accessCheck(FALSE);
        $results = $query->execute();
        $references = array_merge($references, $results);
      }
    }
    return $references;
  }
}
